from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget

__all__ = ['Frame']

@bind(QtWidgets.QFrame)
class Frame(BaseWidget):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .qt import QtCore
from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget

__all__ = ['CheckBox']

@bind(QtWidgets.QCheckBox)
class CheckBox(BaseWidget):

    def __init__(self, text=None, *, checked=False, changed=None, **kwargs):
        super().__init__(**kwargs)
        if text is not None:
            self.text = text
        self.checked = checked
        self.changed = changed
        # Connect signals
        self.qt.stateChanged.connect(self.__handle_changed)

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(value)

    @property
    def checked(self):
        return self.qt.checkState() == QtCore.Qt.Checked

    @checked.setter
    def checked(self, value):
        self.qt.setChecked(QtCore.Qt.Checked if value else QtCore.Qt.Unchecked)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, state):
        if callable(self.changed):
            self.changed(state == QtCore.Qt.Checked)



from .qt import QtWidgets
from .qt import bind

from .widget import Widget

__all__ = ['GroupBox']

@bind(QtWidgets.QGroupBox)
class GroupBox(Widget):

    def __init__(self, title=None, *, checkable=None, checked=None, flat=None,
                 clicked=None, toggled=None, **kwargs):
        super().__init__(**kwargs)
        if title is not None:
            self.title = title
        if checkable is not None:
            self.checkable = checkable
        if checked is not None:
            self.checked = checked
        if flat is not None:
            self.flat = flat
        self.clicked = clicked
        self.toggled = toggled
        # Connect signals
        self.qt.clicked.connect(self.__handle_clicked)
        self.qt.toggled.connect(self.__handle_toggled)

    @property
    def title(self):
        return self.qt.title()

    @title.setter
    def title(self, value):
        self.qt.setTitle(value)

    @property
    def checkable(self):
        return self.qt.isCheckable()

    @checkable.setter
    def checkable(self, value):
        self.qt.setCheckable(value)

    @property
    def checked(self):
        return self.qt.isChecked()

    @checked.setter
    def checked(self, value):
        self.qt.setChecked(value)

    @property
    def flat(self):
        return self.qt.isFlat()

    @flat.setter
    def flat(self, value):
        self.qt.setFlat(value)

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, _):
        if callable(self.clicked):
            self.clicked()

    @property
    def toggled(self):
        return self.__toggled

    @toggled.setter
    def toggled(self, value):
        self.__toggled = value

    def __handle_toggled(self, checked):
        if callable(self.toggled):
            self.toggled(checked)



"""Yet another pythonic UI library for rapid prototyping using
[PyQt5](https://pypi.org/project/PyQt5/).

Qutie (pronounced as _cutie_) provides a simple and easy to use pythonic
interface to [PyQt5](https://pypi.org/project/PyQt5/).

## Quick start

```python
import qutie as ui

app = ui.Application()
window = ui.Widget(
    title="Example",
    icon='orange',
    width=320,
    height=240,
    layout=ui.Column(
        ui.Label("Hello world!", color='blue'),
        ui.Row(
            ui.Button("Click!", clicked=lambda: ui.show_info(text="Hello world!")),
            ui.Button("Quit", clicked=app.quit)
        )
    )
)
window.show()
app.run()
```

## Pythonic API

### Iterables

`list.List`s, `tree.Tree`s and `table.Table`s behave like ```iterators```,
providing ```list```-like methods for access and manipulation.

>>> for item in tree:
...     item[0].color = 'red'
>>> tree[0].value = "spam"
>>> tree.append(["eggs"])
>>> tree.insert(0, ["ham"])
>>> tree.remove(tree[0])
>>> tree.index(tree[0])
0
>>> len(tree)
1

### Events

Any callable class property can be an event, events can emit any kind of
positional and keysword arguments.

>>> widget.my_event = lambda message: print(message)
>>> widget.emit('my_event', "Cheese shop")

### Something missing?

Any underlying PyQt5 instance can be accessed directly using property ```qt```.

>>> widget.qt.setWindowTitle("Spam!")
>>> widget.qt.customContextMenuRequested.conenct(lambda pos: None)

"""

from .action import *
from .application import *
from .button import *
from .checkbox import *
from .combobox import *
from .dialog import *
from .frame import *
from .groupbox import *
from .icon import *
from .label import *
from .layout import *
from .list import *
from .mainwindow import *
from .menu import *
from .messagebox import *
from .number import *
from .object import *
from .pixmap import *
from .progressbar import *
from .scrollarea import *
from .settings import *
from .splitter import *
from .stack import *
from .table import *
from .tabs import *
from .text import *
from .textarea import *
from .timer import *
from .tree import *
from .widget import *
from .worker import *

__version__ = '1.4.1'



import traceback

from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget

__all__ = [
    'MessageBox',
    'show_info',
    'show_warning',
    'show_error',
    'show_exception',
    'show_question'
]

@bind(QtWidgets.QMessageBox)
class MessageBox(BaseWidget):
    """Message box dialog."""

    QtIcons = {
        'information': QtWidgets.QMessageBox.Information,
        'warning': QtWidgets.QMessageBox.Warning,
        'critical': QtWidgets.QMessageBox.Critical,
        'question': QtWidgets.QMessageBox.Question
    }
    QtStandardButtons = {
        'ok': QtWidgets.QMessageBox.Ok,
        'open': QtWidgets.QMessageBox.Open,
        'save': QtWidgets.QMessageBox.Save,
        'cancel': QtWidgets.QMessageBox.Cancel,
        'close': QtWidgets.QMessageBox.Close,
        'discard': QtWidgets.QMessageBox.Discard,
        'apply': QtWidgets.QMessageBox.Apply,
        'reset': QtWidgets.QMessageBox.Reset,
        'restore_defaults': QtWidgets.QMessageBox.RestoreDefaults,
        'help': QtWidgets.QMessageBox.Help,
        'save_all': QtWidgets.QMessageBox.SaveAll,
        'yes': QtWidgets.QMessageBox.Yes,
        'yes_to_all': QtWidgets.QMessageBox.YesToAll,
        'no': QtWidgets.QMessageBox.No,
        'no_to_all': QtWidgets.QMessageBox.NoToAll,
        'abort': QtWidgets.QMessageBox.Abort,
        'retry': QtWidgets.QMessageBox.Retry,
        'ignore': QtWidgets.QMessageBox.Ignore
    }

    def __init__(self, *, icon=None, title=None, text=None, details=None,
                 informative=None, buttons=None, width=None, **kwargs):
        super().__init__(**kwargs)
        if icon is not None:
            self.icon = icon
        if title is not None:
            self.title = title
        if text is not None:
            self.text = text
        if details is not None:
            self.details = details
        if informative is not None:
            self.informative = informative
        if buttons is not None:
            self.buttons = buttons
        # Workaround to resize message box
        layout = self.qt.layout()
        rows = layout.rowCount()
        columns = layout.columnCount()
        self._spacer_item = QtWidgets.QSpacerItem(420, 0)
        layout.addItem(self._spacer_item, rows, 0, 1, columns)
        if width is not None:
            self.width = width

    @property
    def icon(self):
        icon = self.qt.icon()
        for key, value in self.QtIcons.items():
            if value == icon:
                return key
        return None

    @icon.setter
    def icon(self, value):
        self.qt.setIcon(self.QtIcons.get(value, QtWidgets.QMessageBox.NoIcon))

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(value)

    @property
    def details(self):
        return self.qt.detailedText()

    @details.setter
    def details(self, value):
        self.qt.setDetailedText(value)

    @property
    def informative(self):
        return self.qt.informativeText()

    @informative.setter
    def informative(self, value):
        self.qt.setInformativeText(value)

    @property
    def buttons(self):
        buttons = []
        values = self.qt.standardButtons()
        for key, mask in self.QtStandardButtons.items():
            if values & mask:
                buttons.append(key)
        return tuple(buttons)

    @buttons.setter
    def buttons(self, value):
        buttons = 0
        for button in value:
            buttons |= self.QtStandardButtons.get(button, 0)
        self.qt.setStandardButtons(buttons)

    @property
    def width(self):
        return self.qt.width()

    @width.setter
    def width(self, value):
        self._spacer_item.changeSize(value, 0)

    def run(self):
        result = self.qt.exec_()
        for key, value in self.QtStandardButtons.items():
            if value == result:
                return key
        return None

def show_info(text, *, title=None, details=None, **kwargs):
    """Show information message box.

    >>> show_info("Info", "NO-body expects the Spanish Inquisition!")
    """
    dialog = MessageBox(
        icon='information',
        title=title,
        text=text,
        details=details,
        **kwargs
    )
    dialog.run()

def show_warning(text, *, title=None, details=None, **kwargs):
    """Show warning message box.

    >>> show_warning("Warning", "NO-body expects the Spanish Inquisition!")
    """
    dialog = MessageBox(
        icon='warning',
        title=title,
        text=text,
        details=details,
        **kwargs
    )
    dialog.run()

def show_error(text, *, title=None, details=None, **kwargs):
    """Show warning message box.

    >>> show_error("Error", "NO-body expects the Spanish Inquisition!")
    """
    dialog = MessageBox(
        icon='critical',
        title=title,
        text=text,
        details=details,
        **kwargs
    )
    dialog.run()

def show_exception(exception, tb=None, **kwargs):
    """Show exception message box including exception stack trace.

    >>> try:
    ...     foo()
    ... except NameError as e:
    ...     show_exception(e)
    """
    if not tb:
        tb = traceback.format_exc() or None
    show_error(
        title="An exception occured",
        text=format(exception),
        details=tb,
        **kwargs
    )

def show_question(text, *, title=None, details=None, **kwargs):
    """Show question message box, returns True for yes and False for no.

    >>> show_question("Question", "Fancy a cup of Yorkshire Tea?")
    True
    """
    dialog = MessageBox(
        icon='question',
        title=title,
        text=text,
        details=details,
        buttons=('yes', 'no'),
        **kwargs
    )
    return dialog.run() == 'yes'



__all__ = [
    'Base',
]

class Base:
    """Base class used for abstracting Qt5 classes."""

    QtClass = NotImplemented
    """Underlying Qt5 class."""

    def __init__(self, *args):
        self.__qt = self.QtClass(*args)

    @property
    def qt(self):
        """Access underlying Qt5 instance."""
        return self.__qt



from .qt import QtGui
from .qt import QtWidgets
from .qt import bind

from .icon import Icon
from .widget import Widget

__all__ = ['Button', 'RadioButton']

@bind(QtWidgets.QAbstractButton)
class BaseButton(Widget):

    def __init__(self, text=None, *, checkable=None, checked=None, icon=None,
                 clicked=None, toggled=None, pressed=None, released=None,
                 **kwargs):
        super().__init__(**kwargs)
        if text is not None:
            self.text = text
        if checkable is not None:
            self.checkable = checkable
        if checked is not None:
            self.checked = checked
        if icon is not None:
            self.icon = icon
        self.clicked = clicked
        self.pressed = pressed
        self.released = released
        self.toggled = toggled
        # Connect signals
        self.qt.clicked.connect(self.__handle_clicked)
        self.qt.pressed.connect(self.__handle_pressed)
        self.qt.released.connect(self.__handle_released)
        self.qt.toggled.connect(self.__handle_toggled)

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(value)

    @property
    def checkable(self):
        return self.qt.isCheckable()

    @checkable.setter
    def checkable(self, value):
        self.qt.setCheckable(value)

    @property
    def checked(self):
        return self.qt.isChecked()

    @checked.setter
    def checked(self, value):
        self.qt.setChecked(value)

    @property
    def icon(self):
        icon = self.qt.icon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setIcon(value.qt)

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, _):
        if callable(self.clicked):
            self.clicked()

    @property
    def pressed(self):
        return self.__pressed

    @pressed.setter
    def pressed(self, value):
        self.__pressed = value

    def __handle_pressed(self):
        if callable(self.pressed):
            self.pressed()

    @property
    def released(self):
        return self.__released

    @released.setter
    def released(self, value):
        self.__released = value

    def __handle_released(self):
        if callable(self.released):
            self.released()

    @property
    def toggled(self):
        return self.__toggled

    @toggled.setter
    def toggled(self, value):
        self.__toggled = value

    def __handle_toggled(self, checked):
        if callable(self.toggled):
            self.toggled(checked)

    def click(self):
        self.qt.click()

@bind(QtWidgets.QPushButton)
class Button(BaseButton):

    def __init__(self, text=None, *, default=False, auto_default=False,
                 flat=None, **kwargs):
        super().__init__(text, **kwargs)
        if default is not None:
            self.default = default
        if auto_default is not None:
            self.auto_default = auto_default
        if flat is not None:
            self.flat = flat

    @property
    def default(self):
        return self.qt.isDefault()

    @default.setter
    def default(self, value):
        self.qt.setDefault(value)

    @property
    def auto_default(self):
        return self.qt.isAutoDefault()

    @auto_default.setter
    def auto_default(self, value):
        self.qt.setAutoDefault(value)

    @property
    def flat(self):
        return self.qt.isFlat()

    @flat.setter
    def flat(self, value):
        self.qt.setFlat(value)

@bind(QtWidgets.QRadioButton)
class RadioButton(BaseButton):

    pass



from .qt import QtGui
from .qt import bind

from .base import Base
from .pixmap import Pixmap

__all__ = ['Icon']

@bind(QtGui.QIcon)
class Icon(Base):
    """Icon containing multiple resolution pixmaps.

    Load multiple resoultion icons.
    >>> icon = Icon('small.png', 'large.png')

    Load icon from theme (X11 only).
    >>> icon = Icon.from_theme('document-open', Icon('fallback.png'))

    Create a color icon.
    >>> icon = Icon.from_color('red')

    Create a color icon of custom size.
    >>> icon = Icon.from_color('green', 32, 32)
    """

    def __init__(self, *values, qt=None):
        if qt is None:
            super().__init__()
        else:
            super().__init__(qt)
        for value in values:
            self.append(value)

    @property
    def available_sizes(self):
        sizes = []
        for size in self.qt.availableSizes():
            sizes.append((size.width(), size.height()))
        return tuple(sizes)

    @property
    def name(self):
        return self.qt.name()

    @property
    def theme_name(self):
        return self.QtClass.themeName()

    def append(self, value):
        if isinstance(value, str):
            # Try to create color from string
            color = QtGui.QColor(value)
            if color.isValid():
                pixmap = Pixmap.QtClass(64, 64)
                pixmap.fill(color)
                self.qt.addPixmap(pixmap)
            else:
                self.qt.addFile(value)
        elif isinstance(value, Pixmap):
            self.qt.addPixmap(value.qt)
        else:
            raise ValueError(value)

    def pixmap(self, width, height):
        return Pixmap(qt=self.qt.pixmap(width, height))

    @classmethod
    def from_color(cls, color, width=16, height=16):
        color = QtGui.QColor(color)
        pixmap = Pixmap.QtClass(width, height)
        pixmap.fill(color)
        return Icon(qt=pixmap)

    @classmethod
    def from_theme(cls, name, fallback=None):
        args = [name]
        if fallback is not None:
            args.append(fallback)
        icon = cls.QtClass.fromTheme(*args)
        return Icon(qt=icon)



"""Progress bar.

For more information on the underlying Qt5 object see [QProgressBar](https://doc.qt.io/qt-5/qprogressbar.html).
"""

from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget
from .mixins import OrientationMixin

__all__ = ['ProgressBar']

@bind(QtWidgets.QProgressBar)
class ProgressBar(BaseWidget, OrientationMixin):

    def __init__(self, value=0, *, minimum=0, maximum=100, format=None,
                 inverted_appearance=False, orientation=None,
                 text_visible=None, value_changed=None, **kwargs):
        super().__init__(**kwargs)
        self.minimum = minimum
        self.maximum = maximum
        if format is not None:
            self.format = format
        self.inverted_appearance = inverted_appearance
        if orientation is not None:
            self.orientation = orientation
        if text_visible is not None:
            self.text_visible = text_visible
        self.value = value
        self.value_changed = value_changed
        # Connect signals
        self.qt.valueChanged.connect(self.__handle_value_changed)

    @property
    def value(self):
        return self.qt.value()

    @value.setter
    def value(self, value):
        self.qt.setValue(value)

    @property
    def minimum(self):
        return self.qt.minimum()

    @minimum.setter
    def minimum(self, value):
        self.qt.setMinimum(value)

    @property
    def maximum(self):
        return self.qt.maximum()

    @maximum.setter
    def maximum(self, value):
        self.qt.setMaximum(value)

    @property
    def format(self):
        """Format for displayed text, placeholders are %p, %v and %m."""
        return self.qt.format()

    @format.setter
    def format(self, value):
        self.qt.setFormat(value)

    def reset_format(self):
        self.qt.resetFormat()

    @property
    def inverted_appearance(self):
        return self.qt.invertedAppearance()

    @inverted_appearance.setter
    def inverted_appearance(self, value):
        self.qt.setInvertedAppearance(value)

    @property
    def text(self):
        return self.qt.text()

    @property
    def text_visible(self):
        return self.qt.isTextVisible()

    @text_visible.setter
    def text_visible(self, value):
        self.qt.setTextVisible(value)

    @property
    def range(self):
        return self.minimum, self.maximum

    @range.setter
    def range(self, value):
        self.minimum = value[0]
        self.maximum = value[1]

    @property
    def value_changed(self):
        return self.__value_changed

    @value_changed.setter
    def value_changed(self, value):
        self.__value_changed = value

    def __handle_value_changed(self, value):
        if callable(self.value_changed):
            self.value_changed(value)

    def reset(self):
        self.qt.reset()



from .qt import QtWidgets
from .qt import bind

from .action import Action
from .widget import BaseWidget

__all__ = ['Menu']

@bind(QtWidgets.QMenu)
class Menu(BaseWidget):

    def __init__(self, *items, text=None, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)
        if text is not None:
            self.text = text

    @property
    def text(self):
        return self.qt.title()

    @text.setter
    def text(self, value):
        self.qt.setTitle(value)

    def append(self, item):
        if isinstance(item, Action):
            self.qt.addAction(item.qt)
        elif isinstance(item, Menu):
            self.qt.addMenu(item.qt)
        elif isinstance(item, str):
            item = Action(item)
            self.qt.addAction(item.qt)
        else:
            raise ValueError(item)
        return item

    def insert(self, before, item):
        if isinstance(item, Action):
            if isinstance(before, Menu):
                self.qt.insertAction(before.qt.menuAction(), item.qt)
            else:
                self.qt.insertAction(before.qt, item.qt)
        elif isinstance(item, Menu):
            if isinstance(before, Menu):
                self.qt.insertMenu(before.qt.menuAction(), item.qt)
            else:
                self.qt.insertMenu(before.qt, item.qt)
        elif isinstance(item, str):
            item = Action(item)
            if isinstance(before, Menu):
                self.qt.insertAction(before.qt.menuAction(), item.qt)
            else:
                self.qt.insertAction(before.qt, item.qt)
        else:
            raise ValueError(item)
        return item

    def index(self, item):
        return self.qt.actions().index(item)

    def __getitem__(self, index):
        item = self.qt.actions()[index]
        return item.property(self.QtPropertyKey)

    def __iter__(self):
        return iter(item.property(self.QtPropertyKey) for item in self.qt.actions())

    def __len__(self):
        return len(self.qt.actions())



from .qt import QtCore
from .qt import bind

from .base import Base

__all__ = ['Object']

@bind(QtCore.QObject)
class Object(Base):

    QtPropertyKey = '__qutie_ref'

    def __init__(self, *args, object_name=None, destroyed=None,
                 object_name_changed=None):
        super().__init__(*args)
        self.qt.setProperty(self.QtPropertyKey, self)
        if object_name is not None:
            self.object_name = object_name
        self.destroyed = destroyed
        self.object_name_changed = object_name_changed
        # Connect signals
        self.qt.destroyed.connect(self.__handle_destroyed)
        self.qt.objectNameChanged.connect(self.__handle_object_name_changed)
        self.qt.eventEmitted.connect(self.__handle_event)

    @property
    def object_name(self) -> str:
        return self.qt.objectName()

    @object_name.setter
    def object_name(self, value: str):
        self.qt.setObjectName(value)

    @property
    def destroyed(self) -> object:
        return self.__destroyed

    @destroyed.setter
    def destroyed(self, value: object):
        self.__destroyed = value

    def __handle_destroyed(self, obj: object):
        if callable(self.destroyed):
            self.destroyed(obj.property(self.QtPropertyKey))

    @property
    def object_name_changed(self) -> object:
        return self.__object_name_changed

    @object_name_changed.setter
    def object_name_changed(self, value: object):
        self.__object_name_changed = value

    def __handle_object_name_changed(self):
        if callable(self.object_name_changed):
            self.object_name_changed(self.object_name)

    def __handle_event(self, name: str, args: list, kwargs: dict):
        if hasattr(self, name):
            attr = getattr(self, name)
            if callable(attr):
                attr(*args, **kwargs)

    def emit(self, *args, **kwargs):
        """Emit an event.

        >>> o.event = lambda value: print(value) # assign event callback
        >>> o.emit('event', 42)
        """
        if not args:
            raise ValueError("Missing event argument.")
        self.qt.eventEmitted.emit(args[0], args[1:], kwargs)



from .qt import QtWidgets
from .qt import bind

from .menu import Menu
from .widget import BaseWidget

__all__ = ['MainWindow']

@bind(QtWidgets.QMenuBar)
class MenuBar(BaseWidget):

    def __init__(self, *items, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)

    def append(self, item):
        if isinstance(item, str):
            item = Menu(text=item)
        self.qt.addMenu(item.qt)
        return item

    def insert(self, before, item):
        if isinstance(item, str):
            item = Menu(text=item)
        if isinstance(before, Menu):
            self.qt.insertMenu(before.qt.menuAction(), item.qt)
        else:
            self.qt.insertMenu(before.qt, item.qt)
        return item

    def __getitem__(self, index):
        item = self.qt.actions()[index]
        return item.property(self.QtPropertyKey)

    def __iter__(self):
        return iter(item.property(self.QtPropertyKey) for item in self.qt.actions())

    def __len__(self):
        return len(self.qt.actions())

@bind(QtWidgets.QStatusBar)
class StatusBar(BaseWidget):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def append(self, widget):
        self.qt.addPermanentWidget(widget.qt)
        return widget

@bind(QtWidgets.QMainWindow)
class MainWindow(BaseWidget):

    def __init__(self, *, layout=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setMenuBar(MenuBar().qt)
        self.qt.setStatusBar(StatusBar().qt)
        self.layout = layout

    @property
    def layout(self):
        widget = self.qt.centralWidget()
        if widget is not None:
            return widget.property(self.QtPropertyKey)
        return None

    @layout.setter
    def layout(self, value):
        if value is None:
            self.qt.setCentralWidget(None)
        else:
            if not isinstance(value, BaseWidget):
                raise ValueError(value)
            self.qt.setCentralWidget(value.qt)

    @property
    def menubar(self):
        return self.qt.menuBar().property(self.QtPropertyKey)

    @property
    def statusbar(self):
        return self.qt.statusBar().property(self.QtPropertyKey)



"""Stack widget.

For more information on the underlying Qt5 object see [QStackedWidget](https://doc.qt.io/qt-5/qstackedwidget.html).
"""

from .qt import QtCore
from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget

__all__ = ['Stack']

@bind(QtWidgets.QStackedWidget)
class Stack(BaseWidget):

    def __init__(self, *items, changed=None, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)
        self.changed = changed
        # Connect signals
        self.qt.currentChanged.connect(self.__handle_changed)

    @property
    def items(self):
        return list(self)

    def append(self, item):
        self.qt.addWidget(item.qt)

    def insert(self, index, item):
        if index < 0:
            index = max(0, len(self) + index)
        self.qt.insertWidget(index, item.qt)

    def remove(self, item):
        self.qt.removeWidget(item.qt)

    @property
    def current(self):
        item = self.qt.currentIndex()
        if item is not None:
            return item.property(self.QtPropertyKey)
        return item

    @current.setter
    def current(self, item):
        index = self.qt.indexOf(item.qt)
        self.qt.setCurrentIndex(index)

    def index(self, item):
        index = self.qt.indexOf(item.qt)
        if index < 0:
            raise ValueError("item not in stack")
        return index

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, changed):
        self.__changed = changed

    def __handle_changed(self, index):
        if callable(self.changed):
            self.changed(index)

    def __getitem__(self, key):
        item = self.qt.widget(key)
        if not item:
            raise KeyError(key)
        return item.property(self.QtPropertyKey)

    def __setitem__(self, key, value):
        del self[key]
        self.insert(key, value)

    def __delitem__(self, key):
        item = self[key]
        self.qt.removeWidget(item)

    def __len__(self):
        return self.qt.count()

    def __iter__(self):
        return (self[index] for index in range(len(self)))



"""Timer.

For more information on the underlying Qt5 object see [QTimer](https://doc.qt.io/qt-5/qtimer.html).
"""

from .qt import QtCore
from .qt import bind

from .base import Base

__all__ = ['Timer', 'single_shot']

def single_shot(interval, timeout):
    QtCore.QTimer.singleShot(interval * 1e3, timeout)

@bind(QtCore.QTimer)
class Timer(Base):
    """Timer.

    >>> timer = Timer(interval=1.0, timeout=lambda: print("Done!"))
    >>> timer.start()
    """

    QtTimerTypes = {
        'precise': QtCore.Qt.PreciseTimer,
        'coarse': QtCore.Qt.CoarseTimer,
        'very_coarse': QtCore.Qt.VeryCoarseTimer
    }

    def __init__(self, *, interval=0, single_shot=False, timer_type=None,
                 timeout=None, **kwargs):
        super().__init__(**kwargs)
        self.interval = interval
        self.single_shot = single_shot
        if timer_type is not None:
            self.timer_type = timer_type
        self.timeout = timeout
        # Connect signals
        self.qt.timeout.connect(self.__handle_timeout)

    @property
    def interval(self):
        return self.qt.interval() / 1e3

    @interval.setter
    def interval(self, value):
        self.qt.setInterval(value * 1e3)

    @property
    def single_shot(self):
        return self.qt.isSingleShot()

    @single_shot.setter
    def single_shot(self, value):
        self.qt.setSingleShot(value)

    @property
    def timer_type(self):
        index = list(self.QtTimerTypes.values()).index(self.qt.timerType())
        return list(self.QtTimerTypes.keys())[index]

    @timer_type.setter
    def timer_type(self, value):
        self.qt.setTimerType(self.QtTimerTypes[value])

    @property
    def timeout(self):
        """Event called on timer timeout."""
        return self.__timeout

    @timeout.setter
    def timeout(self, value):
        self.__timeout = value

    def __handle_timeout(self):
        if callable(self.timeout):
            self.timeout()

    @property
    def active(self):
        return self.qt.isActive()

    @property
    def remaining_time(self):
        value = self.qt.remainingTime()
        if value < 0:
            return 0.
        return value * 1e3

    @property
    def timer_id(self):
        return self.qt.itmerId()

    def start(self, interval=None):
        if interval is None:
            self.qt.start()
        else:
            self.qt.start(interval * 1e3)

    def stop(self):
        self.qt.stop()



from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

__all__ = [
    'QtCore',
    'QtGui',
    'QtWidgets',
    'bind'
]

def bind(qt):
    """Decorator used to bind custom events on a Qt base class."""
    def bind(cls):
        class QtClass(qt):
            eventEmitted = QtCore.pyqtSignal(str, object, object)
        cls.QtClass = QtClass
        return cls
    return bind



"""Text area input.

For more information on the underlying Qt5 object see [QTextEdit](https://doc.qt.io/qt-5/qtextedit.html).
"""

from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget

__all__ = ['TextArea']

@bind(QtWidgets.QTextEdit)
class TextArea(BaseWidget):
    """Text area widget.

    >>> textarea = TextArea('lorem')
    >>> textarea.append(' ispum')
    >>> textarea.value
    'lorem ipsum'
    >>> tree.clear()
    textarea.value
    ''
    """

    def __init__(self, value=None, *, readonly=False, richtext=False, **kwargs):
        super().__init__(**kwargs)
        self.readonly = readonly
        self.richtext = richtext
        if value is not None:
            self.value = value

    @property
    def value(self):
        return self.qt.toPlainText()

    @value.setter
    def value(self, value):
        self.qt.setPlainText(value)

    @property
    def readonly(self):
        """Read only enabled.

        >>> textarea.readonly = True
        >>> textarea.readonly
        True
        """
        return self.qt.isReadOnly()

    @readonly.setter
    def readonly(self, value):
        self.qt.setReadOnly(value)

    @property
    def richtext(self):
        """Rich Text enabled.

        >>> textarea.richtext = True
        >>> textarea.richtext
        True
        """
        return self.qt.acceptRichText()

    @richtext.setter
    def richtext(self, value):
        self.qt.setAcceptRichText(value)

    def append(self, text):
        """Append text as paragraph to current value.

        >>> textarea = TextArea('lorem')
        >>> textarea.append(' ipsum')
        >>> textarea.value
        'lorem\n ipsum'
        """
        self.qt.append(text)

    def insert(self, text):
        """Insert text at cursor position.

        >>> textarea.insert('ipsum')
        """
        self.qt.insertPlainText(text)

    def clear(self):
        """Clear current value.

        >>> textarea.clear()
        >>> textarea.value
        ''
        """
        self.qt.clear()

    def select_all(self):
        """Select entire text."""
        self.qt.selectAll()

    def redo(self):
        """Redo last user changes."""
        self.qt.redo()

    def undo(self):
        """Undo last user changes."""
        self.qt.undo()



from .qt import QtGui
from .qt import QtWidgets
from .qt import bind

from .object import Object

__all__ = ['Action']

@bind(QtWidgets.QAction)
class Action(Object):

    def __init__(self, text=None, *, checkable=None, checked=False, tool_tip=None,
                 status_tip=None, shortcut=None, separator=None, triggered=None,
                 toggled=None, **kwargs):
        super().__init__(**kwargs)
        if text is not None:
            self.text = text
        if checkable is not None:
            self.checkable = checkable
        self.checked = checked
        if tool_tip is not None:
            self.tool_tip = tool_tip
        if status_tip is not None:
            self.status_tip = status_tip
        if shortcut is not None:
            self.shortcut = shortcut
        if separator is not None:
            self.separator = separator
        self.triggered = triggered
        self.toggled = toggled
        # Connect signals
        self.qt.triggered.connect(self.__handle_triggered)
        self.qt.toggled.connect(self.__handle_toggled)

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(value)

    @property
    def checked(self):
        return self.qt.isChecked()

    @checked.setter
    def checked(self, value):
        self.qt.setChecked(value)

    @property
    def checkable(self):
        return self.qt.isCheckable()

    @checkable.setter
    def checkable(self, value):
        self.qt.setCheckable(value)

    @property
    def tool_tip(self):
        return self.qt.toolTip()

    @tool_tip.setter
    def tool_tip(self, value):
        self.qt.setToolTip(value)

    @property
    def status_tip(self):
        return self.qt.statusTip()

    @status_tip.setter
    def status_tip(self, value):
        self.qt.setStatusTip(value)

    @property
    def shortcut(self):
        return self.qt.shortcut().toString() or None

    @shortcut.setter
    def shortcut(self, value):
        if value is None:
            self.qt.setShortcut(QtGui.QKeySequence())
        else:
            self.qt.setShortcut(QtGui.QKeySequence(value))

    @property
    def separator(self):
        return self.qt.isSeparator()

    @separator.setter
    def separator(self, value):
        self.qt.setSeparator(value)

    @property
    def triggered(self):
        return self.__triggered

    @triggered.setter
    def triggered(self, value):
        self.__triggered = value

    def __handle_triggered(self):
        if callable(self.triggered):
            self.triggered()

    @property
    def toggled(self):
        return self.__toggled

    @toggled.setter
    def toggled(self, value):
        self.__toggled = value

    def __handle_toggled(self, checked):
        if callable(self.toggled):
            self.toggled(checked)

    def trigger(self):
        self.qt.trigger()

    def toggle(self):
        self.qt.toggle()



from .qt import QtWidgets
from .qt import bind

from .widget import BaseWidget
from .widget import Widget
from .mixins import OrientationMixin

__all__ = [
    'Dialog',
    'DialogButtonBox',
    'filename_open',
    'filenames_open',
    'directory_open',
    'filename_save',
    'get_number',
    'get_text',
    'get_item'
]

@bind(QtWidgets.QDialog)
class Dialog(Widget):

    def __init__(self, *, accepted=None, rejected=None, **kwargs):
        super().__init__(**kwargs)

        self.accepted = accepted
        def accepted_event():
            if callable(self.accepted):
                self.accepted()
        self.qt.accepted.connect(accepted_event)

        self.rejected = rejected
        def rejected_event():
            if callable(self.rejected):
                self.rejected()
        self.qt.rejected.connect(rejected_event)

    def accept(self):
        return self.qt.accept()

    def reject(self):
        return self.qt.reject()

    def run(self):
        return {
            self.qt.Accepted: True,
            self.qt.Rejected: False
        }[self.qt.exec_()]

@bind(QtWidgets.QDialogButtonBox)
class DialogButtonBox(BaseWidget, OrientationMixin):

    QtStandardButtons = {
        'ok': QtWidgets.QDialogButtonBox.Ok,
        'open': QtWidgets.QDialogButtonBox.Open,
        'save': QtWidgets.QDialogButtonBox.Save,
        'cancel': QtWidgets.QDialogButtonBox.Cancel,
        'close': QtWidgets.QDialogButtonBox.Close,
        'discard': QtWidgets.QDialogButtonBox.Discard,
        'apply': QtWidgets.QDialogButtonBox.Apply,
        'reset': QtWidgets.QDialogButtonBox.Reset,
        'restore_defaults': QtWidgets.QDialogButtonBox.RestoreDefaults,
        'help': QtWidgets.QDialogButtonBox.Help,
        'save_all': QtWidgets.QDialogButtonBox.SaveAll,
        'yes': QtWidgets.QDialogButtonBox.Yes,
        'yes_to_all': QtWidgets.QDialogButtonBox.YesToAll,
        'no': QtWidgets.QDialogButtonBox.No,
        'no_to_all': QtWidgets.QDialogButtonBox.NoToAll,
        'abort': QtWidgets.QDialogButtonBox.Abort,
        'retry': QtWidgets.QDialogButtonBox.Retry,
        'ignore': QtWidgets.QDialogButtonBox.Ignore
    }

    def __init__(self, buttons=None, orientation=None, accepted=None,
                 rejected=None, clicked=None, help_requested=None, **kwargs):
        super().__init__(**kwargs)
        if buttons is not None:
            self.buttons = buttons
        if orientation is not None:
            self.orientation = orientation
        self.accepted = accepted
        self.rejected = rejected
        self.clicked = clicked
        self.help_requested = help_requested
        # Connect signals
        self.qt.accepted.connect(self.__handle_accepted)
        self.qt.rejected.connect(self.__handle_rejected)
        self.qt.clicked.connect(self.__handle_clicked)
        self.qt.helpRequested.connect(self.__handle_help_requested)

    @property
    def buttons(self):
        buttons = []
        values = self.qt.standardButtons()
        for key, mask in self.QtStandardButtons.items():
            if values & mask:
                buttons.append(key)
        return tuple(buttons)

    @buttons.setter
    def buttons(self, value):
        buttons = 0
        for button in value:
            buttons |= self.QtStandardButtons.get(button, 0)
        self.qt.setStandardButtons(buttons)

    @property
    def accepted(self):
        return self.__accepted

    @accepted.setter
    def accepted(self, value):
        self.__accepted = value

    def __handle_accepted(self):
        if callable(self.accepted):
            self.accepted()

    @property
    def rejected(self):
        return self.__rejected

    @rejected.setter
    def rejected(self, value):
        self.__rejected = value

    def __handle_rejected(self):
        if callable(self.rejected):
            self.rejected()

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, button):
        if callable(self.clicked):
            button = self.qt.standardButton(button)
            for key, value in self.QtStandardButtons.items():
                if button == value:
                    self.clicked(key)
                    break

    @property
    def help_requested(self):
        return self.__help_requested

    @help_requested.setter
    def help_requested(self, value):
        self.__help_requested = value

    def __handle_help_requested(self):
        if callable(self.help_requested):
            self.help_requested()

def filename_open(path=None, *, filter=None, title=None):
    """Shows a filename selection dialog, returns selected filename path.

    >>> filename_open("/home/user", filter="Text (*.txt)")
    '/home/user/example.txt'
    """
    if title is None:
        title = "Open file"
    return QtWidgets.QFileDialog.getOpenFileName(None, title, path, filter)[0] or None

def filenames_open(path=None, *, filter=None, title=None):
    """Shows a multiple filenames selection dialog, returns list of selected
    filename paths.

    >>> filename_open("/home/user", filter="Text (*.txt)")
    ['/home/user/example.txt', '/home/user/another.txt']
    """
    if title is None:
        title = "Open files"
    return QtWidgets.QFileDialog.getOpenFileNames(None, title, path, filter)[0] or None

def directory_open(path=None, *, title=None):
    """Shows a multiple filenames selection dialog, returns selected directory
    path.

    >>> filename_open(""/home/user")
    '/tmp'
    """
    return QtWidgets.QFileDialog.getExistingDirectory(None, title or "Open directory", path) or None

def filename_save(path=None, *, filter=None, title=None):
    """Shows a save filename selection dialog, returns selected filename path.

    >>> filename_save("/home/user/example.txt", filter="Text (*.txt)")
    '/home/user/other.txt'
    """
    if title is None:
        title = "Save file"
    return QtWidgets.QFileDialog.getSaveFileName(None, title, path, filter)[0] or None

def get_number(value=0, *, minimum=None, maximum=None, decimals=0, title=None, label=None):
    """Number input dialog, optionally editable.

    >>> get_number(value=4.2, minimum=0, maximum=10, decimals=1)
    'Pear'
    """
    value = float(value)
    if minimum is None:
        minimum = float('-inf')
    if maximum is None:
        maximum = float('-inf')
    if title is None:
        title = ""
    if label is None:
        label = ""
    value, success = QtWidgets.QInputDialog.getDouble(
        None, title, label, value, minimum, maximum, decimals
    )
    if success:
        return value
    return None

def get_text(text=None, *, title=None, label=None):
    """Text input dialog.

    >>> get_text()
    'Pear'
    """
    text = format(text)
    if title is None:
        title = ""
    if label is None:
        label = ""
    value, success = QtWidgets.QInputDialog.getText(
        None, title, label, QtWidgets.QLineEdit.Normal, text
    )
    if success:
        return value
    return None

def get_item(items, *, current=0, editable=False, title=None, label=None):
    """Select item from input list dialog, optionally editable.

    >>> get_item(["Apple", "Pear", "Nut"])
    'Pear'
    """
    items = [format(item) for item in items]
    if title is None:
        title = ""
    if label is None:
        label = ""
    item, success = QtWidgets.QInputDialog.getItem(
        None, title, label, items, current, editable
    )
    if success:
        return item
    return None



from .qt import QtCore

__all__ = ['OrientationMixin']

class OrientationMixin:

    horizontal = 'horizontal'
    vertical = 'vertical'

    @property
    def orientation(self):
        return {
            QtCore.Qt.Horizontal: OrientationMixin.horizontal,
            QtCore.Qt.Vertical: OrientationMixin.vertical
        }[self.qt.orientation()]

    @orientation.setter
    def orientation(self, value):
        self.qt.setOrientation({
            OrientationMixin.horizontal: QtCore.Qt.Horizontal,
            OrientationMixin.vertical: QtCore.Qt.Vertical
        }[value])



from .qt import QtGui
from .qt import bind

from .base import Base

__all__ = ['Pixmap']

@bind(QtGui.QPixmap)
class Pixmap(Base):

    def __init__(self, filename=None, *, qt=None):
        if qt is None:
            super().__init__()
        else:
            super().__init__(qt)
        if filename is not None:
            self.load(filename)

    @property
    def width(self):
        return self.qt.width()

    @property
    def height(self):
        return self.qt.height()

    @property
    def size(self):
        return self.width, self.height

    def load(self, filename):
        """Loa pixmap from file."""
        return self.qt.load(filename)

    def save(self, filename, quality=-1):
        """Save pixmap to file."""
        return self.qt.save(filename, quality)

    def fill(self, color):
        """Fill pixmap with color."""
        self.qt.fill(QtGui.QColor(color))

    @classmethod
    def create(cls, width, height, color=None):
        """Return new pixmap instance, with optional fill color."""
        pixmap = Pixmap(qt=cls.QtClass(width, height))
        if color is not None:
            pixmap.fill(color)
        return pixmap
